﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace FATX_Browser.FATX
{
    public struct EntryData
    {
        /// <summary>
        /// Size of the name
        /// </summary>
        public byte FileNameSize;
        /// <summary>
        /// Entry flags
        /// </summary>
        public byte Flags;
        /// <summary>
        /// Entry name
        /// </summary>
        public string FileName;
        /// <summary>
        /// Beginning cluster (block) for the entry
        /// </summary>
        public UInt32 StartingCluster;
        /// <summary>
        /// Size of the entry (0x00 for folders)
        /// </summary>
        public UInt32 Size;
        public Int16 CreationDate;
        public Int16 CreationTime;
        public Int16 AccessDate;
        public Int16 AccessTime;
        public Int16 ModifiedDate;
        public Int16 ModifiedTime;
        /// <summary>
        /// Offset for the entry
        /// </summary>
        public long EntryOffset;
    }

    public class Partition
    {
        PartitionInfo partInfo = new PartitionInfo();

        public PartitionInfo PartInfo
        {
            get { return partInfo; }
            set { partInfo = value; }
        }
    }

    public class Entry
    {
        public EntryData EData = new EntryData();
        /// <summary>
        /// Information about the partition in which the file is located
        /// </summary>
        public PartitionInfo PartInfo;
        internal Entries e;
        internal Misc Misc;
        private FATX.FATXDrive drive;
        private uint[] blocksOCCUPADO;
        internal bool isfolder;

        /// <summary>
        /// Device ID (used for the FATXDrive)
        /// </summary>
        public int DeviceID
        {
            get { return Drive.DeviceID; }
        }

        /// <summary>
        /// Name of the folder
        /// </summary>
        public string Name
        {
            get { return EData.FileName; }
            set { EData.FileName = value; }
        }

        /// <summary>
        /// Offset of the first block
        /// </summary>
        public long BaseOffset
        {
            get { return Misc.GetBlockOffset(Block, this); }
        }

        /// <summary>
        /// Root block
        /// </summary>
        public uint Block
        {
            get { return (uint)EData.StartingCluster; }
        }

        /// <summary>
        /// All blocks that the object occupies
        /// </summary>
        public uint[] BlocksOccupied
        {
            get { 
                if (blocksOCCUPADO == null)
                {
                    blocksOCCUPADO = e.GetBlocksOccupied(Block);
                }
                return blocksOCCUPADO;
            }
            set
            {
                blocksOCCUPADO = value;
            }
        }

        public bool BlocksLoaded
        {
            get { if (blocksOCCUPADO == null) { return false; } else { return true; } }
        }

        /// <summary>
        /// Offset for the entry in the file's folder
        /// </summary>
        public long EntryOffset
        {
            get { return EData.EntryOffset; }
        }

        public FATXDrive Drive
        {
            get { return drive; }
            set { drive = value; }
        }

        public bool IsFolder
        {
            get { return isfolder; }
        }
    }
    
    public sealed class Folder : Entry
    {
        private Folder[] subFolders;
        private File[] files;

        public Folder(FATXDrive DRIVE, PartitionInfo partition)
        {
            PartInfo = partition;
            e = new Entries(this);
            Drive = DRIVE;
            Misc = new Misc();
            isfolder = true;
        }

        /// <summary>
        /// Extracts each subfolder and file in this folder recursively
        /// </summary>
        /// <param name="path">Path to extract to</param>
        /// <param name="updateFolder">Current file being extracted</param>
        /// <param name="updateText">Current file being extracted</param>
        /// <param name="progress">The number of blocks that have been read/written</param>
        /// <param name="maxVal">File size</param>
        /// <param name="CreateRoot">Create root folder</param>
        public bool ExtractRecursive(string path, ref string updateFolder, ref string updateText, ref long progress, ref long maxVal, bool CreateRoot)
        {
            bool result = false;
            try
            {
                updateFolder = Name;
                if (CreateRoot == true)
                {
                    System.IO.Directory.CreateDirectory(path + "\\" + Name);
                    path += "\\" + EData.FileName;
                }

                File[] ourFiles = Files();
                if (ourFiles.Length != 0)
                {
                    foreach (File f in ourFiles)
                    {
                        updateText = f.Name;
                        e.ExtractFile(f, path + "\\" + f.Name, ref progress, ref maxVal);
                    }
                }

                Folder[] subFolders = SubFolders();
                if (subFolders.Length != 0)
                {
                    foreach (Folder f in subFolders)
                    {
                        updateFolder = f.Name;
                        result = f.ExtractRecursive(path, ref updateFolder, ref updateText, ref progress, ref maxVal ,true);
                    }
                }
                return true;
            }
            catch (Exception E) { MessageBox.Show(E.Message); return false; }
        }

        /// <summary>
        /// Returns an array of files that this folder holds
        /// </summary>
        public File[] Files()
        {
            if (BlocksOccupied != null)
            {
                if (files == null)
                {
                    files = e.Files(BlocksOccupied);
                }
                return files;
            }
            else { if (files == null) { files = e.Files(Block);} return files; }
        }

        /// <summary>
        /// Returns an array of subfolders that this folder holds
        /// </summary>
        public Folder[] SubFolders()
        {
            if (BlocksOccupied != null)
            {
                if (subFolders == null)
                {
                    subFolders = e.Folders(BlocksOccupied);
                }
                return subFolders;
            }
            else { if (subFolders == null) { subFolders = e.Folders(Block); } return subFolders; }
        }

        public bool InjectFile(string FilePath)
        {
            return true;
        }

        public bool InjectFolder(string FolderPath)
        {
            return true;
        }
    }

    public sealed class File : Entry
    {
        public File(FATXDrive DRIVE, PartitionInfo partition)
        {
            PartInfo = partition;
            e = new Entries(this);
            Drive = DRIVE;
            Misc = new Misc();
        }

        /// <summary>
        /// Size of the folder in bytes
        /// </summary>
        public long Size
        {
            get { return EData.Size; }
        }

        public bool OverWrite()
        {
            return true;
        }
    }
}
